@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2019 - 2024, Ilan Schnell; All Rights Reserved
+   Copyright (c) 2019 - 2025, Ilan Schnell; All Rights Reserved
    bitarray is published under the PSF license.
 
    This file contains the C implementation of some useful utility functions.
@@ -54,7 +54,7 @@ new_bitarray(Py_ssize_t nbits, PyObject *endian, int c)
 
     assert(res->nbits == nbits && res->readonly == 0 && res->buffer == NULL);
     assert(-1 <= c && c < 256);
-    if (c >= 0)
+    if (c >= 0 && nbits)
         memset(res->ob_item, c, (size_t) Py_SIZE(res));
 
     return res;
@@ -92,7 +92,7 @@ PyDoc_STRVAR(zeros_doc,
 "zeros(length, /, endian=None) -> bitarray\n\
 \n\
 Create a bitarray of length, with all values 0, and optional\n\
-endianness, which may be 'big', 'little'.");
+bit-endianness, which may be 'big', 'little'.");
 
 
 static PyObject *
@@ -113,7 +113,7 @@ PyDoc_STRVAR(ones_doc,
 "ones(length, /, endian=None) -> bitarray\n\
 \n\
 Create a bitarray of length, with all values 1, and optional\n\
-endianness, which may be 'big', 'little'.");
+bit-endianness, which may be 'big', 'little'.");
 
 /* ------------------------------- count_n ----------------------------- */
 
@@ -389,8 +389,9 @@ Return tuple with counts of: ~a & ~b, ~a & b, a & ~b, a & b");
 
 /*
   The binary format used here is similar to the one used for pickling
-  bitarray objects.  This format encodes the bit-endianness in the head byte,
-  whereas the binary pickle blob does not.
+  bitarray objects.  However, this format has a head byte which encodes both
+  the bit-endianness and the number of pad bits, whereas the binary pickle
+  blob does not.
 */
 
 static PyObject *
@@ -431,6 +432,7 @@ deserialize(PyObject *module, PyObject *buffer)
     Py_buffer view;
     bitarrayobject *a;
     unsigned char head;
+    Py_ssize_t nbits;
 
     if (PyObject_GetBuffer(buffer, &view, PyBUF_SIMPLE) < 0)
         return NULL;
@@ -448,9 +450,8 @@ deserialize(PyObject *module, PyObject *buffer)
         goto error;
     }
     /* create bitarray of desired length */
-    a = new_bitarray(8 * (view.len - 1) - ((Py_ssize_t) (head & 0x07)),
-                     Py_None, -1);
-    if (a == NULL)
+    nbits = 8 * (view.len - 1) - ((Py_ssize_t) (head & 0x07));
+    if ((a = new_bitarray(nbits, Py_None, -1)) == NULL)
         goto error;
     /* set bit-endianness and buffer */
     a->endian = head & 0x10 ? ENDIAN_BIG : ENDIAN_LITTLE;
@@ -459,6 +460,7 @@ deserialize(PyObject *module, PyObject *buffer)
 
     PyBuffer_Release(&view);
     return (PyObject *) a;
+
  error:
     PyBuffer_Release(&view);
     return NULL;
@@ -501,7 +503,7 @@ ba2hex_core(bitarrayobject *a)
     if (str == NULL)
         return NULL;
 
-    /* translate entire bitarray buffer, even when we have 4 padbits */
+    /* translate entire bitarray buffer, even when we have 4 pad bits */
     for (i = 0; i < strsize; i += 2) {
         unsigned char c = *buff++;
         str[i + le] = hexdigits[c >> 4];
@@ -553,7 +555,9 @@ hex2ba_core(bitarrayobject *a, Py_buffer hexstr)
 
     assert(a->nbits == 4 * hexstr.len);
 
-    memset(a->ob_item, 0, Py_SIZE(a));
+    if (a->ob_item)
+        memset(a->ob_item, 0, Py_SIZE(a));
+
     for (i = 0; i < hexstr.len; i++) {
         unsigned char c = str[i];
         int x = hex_to_int(c);
@@ -590,6 +594,7 @@ hex2ba(PyObject *module, PyObject *args, PyObject *kwds)
 
     PyBuffer_Release(&hexstr);
     return (PyObject *) a;
+
  error:
     PyBuffer_Release(&hexstr);
     Py_XDECREF((PyObject *) a);
@@ -786,6 +791,7 @@ base2ba(PyObject *module, PyObject *args, PyObject *kwds)
 
     PyBuffer_Release(&asciistr);
     return (PyObject *) a;
+
  error:
     PyBuffer_Release(&asciistr);
     Py_XDECREF((PyObject *) a);
@@ -806,27 +812,21 @@ standard base 64 alphabet is used.");
 static int
 resize_lite(bitarrayobject *self, Py_ssize_t nbits)
 {
-    const Py_ssize_t allocated = self->allocated, size = Py_SIZE(self);
-    const Py_ssize_t newsize = BYTES(nbits);
+    const size_t size = Py_SIZE(self);
+    const size_t allocated = self->allocated;
+    const size_t newsize = BYTES((size_t) nbits);
     size_t new_allocated;
 
-    assert(allocated >= size && size == BYTES(self->nbits));
+    assert(allocated >= size && size == BYTES((size_t) self->nbits));
+    assert(self->readonly == 0);
     assert(self->ob_exports == 0);
     assert(self->buffer == NULL);
-    assert(self->readonly == 0);
 
     if (newsize == size) {
         self->nbits = nbits;
         return 0;
     }
 
-    if (allocated >= newsize && newsize >= (allocated >> 1)) {
-        assert(self->ob_item != NULL || newsize == 0);
-        Py_SET_SIZE(self, newsize);
-        self->nbits = nbits;
-        return 0;
-    }
-
     if (newsize == 0) {
         PyMem_Free(self->ob_item);
         self->ob_item = NULL;
@@ -836,13 +836,23 @@ resize_lite(bitarrayobject *self, Py_ssize_t nbits)
         return 0;
     }
 
-    new_allocated = ((size_t) newsize + (newsize >> 4) +
-                     (newsize < 8 ? 3 : 7)) & ~(size_t) 3;
+    if (allocated >= newsize) {
+        if (newsize >= allocated / 2) {
+            Py_SET_SIZE(self, newsize);
+            self->nbits = nbits;
+            return 0;
+        }
+        new_allocated = newsize;
+    }
+    else {
+        new_allocated = newsize;
+        if (size != 0 && newsize / 2 <= allocated) {
+            new_allocated += (newsize >> 4) + (newsize < 8 ? 3 : 7);
+            new_allocated &= ~(size_t) 3;
+        }
+    }
 
-    if (newsize - size > (Py_ssize_t) new_allocated - newsize)
-        new_allocated = ((size_t) newsize + 3) & ~(size_t) 3;
-
-    assert(new_allocated >= (size_t) newsize);
+    assert(new_allocated >= newsize);
     self->ob_item = PyMem_Realloc(self->ob_item, new_allocated);
     if (self->ob_item == NULL) {
         PyErr_NoMemory();
@@ -1044,18 +1054,21 @@ sc_rts(PyObject *module, PyObject *obj)
         return NULL;
 
     if ((list = PyList_New(NSEG(a->nbits) + 1)) == NULL)
-        return NULL;
+        goto error;
 
     for (i = 0; i <= NSEG(a->nbits); i++) {
         PyObject *item = PyLong_FromSsize_t(rts[i]);
-        if (item == NULL) {
-            Py_DECREF(list);
-            return NULL;
-        }
+        if (item == NULL)
+            goto error;
         PyList_SET_ITEM(list, i, item);
     }
     PyMem_Free(rts);
     return list;
+
+ error:
+    Py_XDECREF(list);
+    PyMem_Free(rts);
+    return NULL;
 }
 #endif  /* NDEBUG */
 
@@ -1307,7 +1320,7 @@ sc_encode(PyObject *module, PyObject *obj)
     bitarrayobject *a;
     Py_ssize_t offset = 0;      /* block offset into bitarray a in bytes */
     Py_ssize_t *rts;            /* running totals for 256 bit segments */
-    Py_ssize_t t;               /* total population count of bitarray a */
+    Py_ssize_t total;           /* total population count of bitarray a */
 
     if (ensure_bitarray(obj) < 0)
         return NULL;
@@ -1319,15 +1332,15 @@ sc_encode(PyObject *module, PyObject *obj)
 
     out = PyBytes_FromStringAndSize(NULL, 32768);
     if (out == NULL)
-        return NULL;
+        goto error;
 
     str = PyBytes_AS_STRING(out);
     len += sc_encode_header(str, a);
 
-    t = rts[NSEG(a->nbits)];
+    total = rts[NSEG(a->nbits)];
     /* encode blocks as long as we haven't reached the end of the bitarray
        and haven't reached the total population count yet */
-    while (offset < Py_SIZE(a) && rts[offset / SEGSIZE] != t) {
+    while (offset < Py_SIZE(a) && rts[offset / SEGSIZE] != total) {
         Py_ssize_t allocated;   /* size (in bytes) of output buffer */
 
         /* Make sure we have enough space in output buffer for next block.
@@ -1337,7 +1350,7 @@ sc_encode(PyObject *module, PyObject *obj)
         allocated = PyBytes_GET_SIZE(out);
         if (allocated < len + 1 + 128 * 32 + 1) {  /* increase allocation */
             if (_PyBytes_Resize(&out, allocated + 32768) < 0)
-                return NULL;
+                goto error;
             str = PyBytes_AS_STRING(out);
         }
         offset += sc_encode_block(str, &len, a, rts, offset);
@@ -1349,6 +1362,10 @@ sc_encode(PyObject *module, PyObject *obj)
         return NULL;
 
     return out;
+
+ error:
+    PyMem_Free(rts);
+    return NULL;
 }
 
 PyDoc_STRVAR(sc_encode_doc,
@@ -1586,6 +1603,7 @@ vl_decode(PyObject *module, PyObject *args, PyObject *kwds)
 
     Py_DECREF(iter);
     return (PyObject *) a;
+
  error:
     Py_DECREF(iter);
     Py_XDECREF((PyObject *) a);
@@ -1905,18 +1923,17 @@ PyInit__util(void)
     PyObject *m, *bitarray_module;
 
     if ((bitarray_module = PyImport_ImportModule("bitarray")) == NULL)
-        goto error;
+        return NULL;
     bitarray_type_obj = PyObject_GetAttrString(bitarray_module, "bitarray");
     Py_DECREF(bitarray_module);
     if (bitarray_type_obj == NULL)
-        goto error;
+        return NULL;
 
-    m = PyModule_Create(&moduledef);
-    if (m == NULL)
-        goto error;
+    if ((m = PyModule_Create(&moduledef)) == NULL)
+        return NULL;
 
     if (PyType_Ready(&CHDI_Type) < 0)
-        goto error;
+        return NULL;
     Py_SET_TYPE(&CHDI_Type, &PyType_Type);
 
 #ifndef NDEBUG  /* expose segment size in debug mode for testing */
@@ -1924,6 +1941,4 @@ PyInit__util(void)
 #endif
 
     return m;
- error:
-    return NULL;
 }
