@@ -1,10 +1,10 @@
 /*
-   Copyright (c) 2008 - 2023, Ilan Schnell; All Rights Reserved
+   Copyright (c) 2008 - 2025, Ilan Schnell; All Rights Reserved
    bitarray is published under the PSF license.
 
    Author: Ilan Schnell
 */
-#define BITARRAY_VERSION  "2.9.1"
+#define BITARRAY_VERSION  "3.1.0"
 
 #ifdef STDC_HEADERS
 #  include <stddef.h>
@@ -33,24 +33,9 @@
 #define Py_UNREACHABLE() abort()
 #endif
 
-#if PY_MAJOR_VERSION >= 3
-#define IS_PY3K  1
-#define BYTES_SIZE_FMT  "y#"
-#else
-#define IS_PY3K  0
-/* the Py_MIN and Py_MAX macros were introduced in Python 3.3 */
-#define Py_MIN(x, y)  (((x) > (y)) ? (y) : (x))
-#define Py_MAX(x, y)  (((x) > (y)) ? (x) : (y))
-#define PySlice_GetIndicesEx(slice, len, start, stop, step, slicelength) \
-    PySlice_GetIndicesEx(((PySliceObject *) slice),                      \
-                         (len), (start), (stop), (step), (slicelength))
-#define PyLong_FromLong  PyInt_FromLong
-#define BYTES_SIZE_FMT  "s#"
-#endif
-
 /* --- bitarrayobject --- */
 
-/* .ob_size is buffer size (in bytes), not the number of elements.
+/* .ob_size is the buffer size (in bytes), not the number of elements.
    The number of elements (bits) is .nbits. */
 typedef struct {
     PyObject_VAR_HEAD
@@ -75,9 +60,9 @@ typedef struct {
 #define ENDIAN_STR(endian)  ((endian) == ENDIAN_LITTLE ? "little" : "big")
 
 /* number of pad bits */
-#define PADBITS(self)  (8 * Py_SIZE(self) - (self)->nbits)
+#define PADBITS(self)  ((8 - (self)->nbits % 8) % 8)
 
-/* number of bytes necessary to store given bits */
+/* number of bytes necessary to store given nunmber of bits */
 #define BYTES(bits)  (((bits) + 7) >> 3)
 
 /* we're not using bitmask_table here, as it is actually slower */
@@ -159,7 +144,7 @@ zlw(bitarrayobject *self)       /* zlw = zeroed last word */
     uint64_t res = 0;
 
     assert(nw + nr == nbits / 8 && nw + nr <= Py_SIZE(self));
-    memcpy((char *) &res, self->ob_item + nw, nr);
+    memcpy((char *) &res, self->ob_item + nw, (size_t) nr);
     if (nbits % 8)
         *(((char *) &res) + nr) = zlc(self);
 
@@ -171,10 +156,11 @@ zlw(bitarrayobject *self)       /* zlw = zeroed last word */
 static inline void
 set_padbits(bitarrayobject *self)
 {
-    const int r = self->nbits % 8;     /* index into mask table */
-
-    if (self->readonly == 0 && r)
-        self->ob_item[Py_SIZE(self) - 1] &= ones_table[IS_BE(self)][r];
+    if (self->readonly == 0) {
+        int r = self->nbits % 8;     /* index into mask table */
+        if (r)
+            self->ob_item[Py_SIZE(self) - 1] &= ones_table[IS_BE(self)][r];
+    }
 }
 
 /* population count - number of 1's in uint64 */
@@ -232,13 +218,13 @@ popcnt_words(uint64_t *w, Py_ssize_t n)
 {
     Py_ssize_t cnt = 0;
 
-    assert(n >= 0 && to_aligned((void *) w) == 0);
+    assert(n >= 0 && ((uintptr_t) w) % 4 == 0);
     while (n--)
         cnt += popcnt_64(*w++);
     return cnt;
 }
 
-/* adjust index a manner consistent with the handling of normal slices */
+/* adjust index in a manner consistent with the handling of normal slices */
 static inline void
 adjust_index(Py_ssize_t length, Py_ssize_t *i, Py_ssize_t step)
 {
@@ -257,7 +243,7 @@ static inline Py_ssize_t
 adjust_indices(Py_ssize_t length, Py_ssize_t *start, Py_ssize_t *stop,
                Py_ssize_t step)
 {
-#if PY_VERSION_HEX > 0x03060100
+#if PY_VERSION_HEX >= 0x03060100
     return PySlice_AdjustIndices(length, start, stop, step);
 #else
     assert(step != 0);
